import abc
import copy
import logging
import operator

from functools import partial

# Can't check for cycles statically because references could be nested for
# any reason. The only time to ever discover this would be during actual
# resolution.


def _resolve_sequence(sequence, context, container_coerce, item_coerce=None):
    item_coerce = item_coerce(sequence) if item_coerce else sequence
    return container_coerce(_resolve_item(item, context)
                            for item in item_coerce)


# TODO(clouser) Find some way to make this available to custom resolve
# callbacks.
def _resolve_item(item, context):
    if isinstance(item, BaseRef):
        result = item.resolve(context)
    else:
        result = item

    if type(result) in context.callbacks:
        result = context.callbacks[type(result)](result, context)
    return result


class ResolutionContext(object):

    def __init__(self, container, callbacks):
        self._container = container
        self._resolved_items = {}
        self._path = []
        self._seen_items = set()
        self.callbacks = copy.deepcopy(callbacks)

    def add_callback(self, item_type, callback):
        logging.debug('Overriding existing callback for %s.', item_type)
        self.callbacks[item_type] = callback

    def Get(self, setting):
        if setting in self._resolved_items:
            return self._resolved_items[setting]

        if setting not in self._container:
            raise ValueError(
                "'{name}' could not be resolved in context {keys}.".format(
                    name=setting, keys=sorted(self._container.keys())))

        if setting in self._seen_items:
            raise ValueError('Circular dependency found: '
                             + ' -> '.join(self._path + [setting]))

        self._path.append(setting)
        self._seen_items.add(setting)

        resolved_item = _resolve_item(self._container[setting], self)

        self._seen_items.remove(setting)
        self._path.pop()

        self._resolved_items[setting] = resolved_item
        return resolved_item


def format_call(method, *args, **kwargs):
    formatted_args = ', '.join(repr(arg) for arg in args)
    formatted_kwargs = ', '.join('%s=%r' % (key, value)
                                 for (key, value) in kwargs.iteritems())
    all_args = ', '.join(filter(bool, [formatted_args, formatted_kwargs]))
    return '{func}({args})'.format(func=method.__name__, args=all_args)


def simple_call_formatter(operator_char, method, *args, **unused_kwargs):
    formatted_operator = ' %s ' % operator_char
    return formatted_operator.join(repr(arg) for arg in args)


class BaseRef(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def resolve(self, resolution_context):
        raise NotImplementedError()

    @staticmethod
    def _make_magic_method(func, operator_char, reverse=False):
        def __operator_func(func_self, other):
            args = [func_self, other]
            if reverse:
                args = list(reversed(args))

            call_formatter = partial(simple_call_formatter, operator_char)
            return FormattedCall(call_formatter, func, *args)
        return __operator_func

    __add__ = _make_magic_method.__func__(operator.add, '+')
    __radd__ = _make_magic_method.__func__(operator.add, '+', reverse=True)
    __sub__ = _make_magic_method.__func__(operator.sub, '-')
    __rsub__ = _make_magic_method.__func__(operator.sub, '-', reverse=True)
    __mul__ = _make_magic_method.__func__(operator.mul, '*')
    __rmul__ = _make_magic_method.__func__(operator.mul, '*', reverse=True)


class Ref(BaseRef):

    def __init__(self, name):
        self._name = name

    def resolve(self, resolution_context):
        return resolution_context.Get(self._name)


class FormattedCall(BaseRef):

    def __init__(self, call_formatter, func, *args, **kwargs):
        self.call_formatter = call_formatter
        self.func = func
        self.args = args
        self.kwargs = kwargs

    def resolve(self, resolution_context):
        resolved_args = _resolve_item(self.args, resolution_context)
        resolved_kwargs = _resolve_item(self.kwargs, resolution_context)
        try:
            return self.func(*resolved_args, **resolved_kwargs)
        except Exception, e:
            formatted_call = self.call_formatter(self.func, *resolved_args,
                                                 **resolved_kwargs)
            raise TypeError(('Problem resolving call: {call}. '
                             'Inner exception: {inner}').format(
                            call=formatted_call,
                            inner=e))


class Call(BaseRef):

    def __init__(self, func, *args, **kwargs):
        self.func = func
        self.args = args
        self.kwargs = kwargs

    def resolve(self, resolution_context):
        resolved_args = _resolve_item(self.args, resolution_context)
        resolved_kwargs = _resolve_item(self.kwargs, resolution_context)
        try:
            return self.func(*resolved_args, **resolved_kwargs)
        except Exception, e:
            formatted_call = format_call(self.func, *resolved_args,
                                         **resolved_kwargs)
            raise TypeError(('Problem resolving call: {call}. '
                             'Inner exception: {inner}').format(
                            call=formatted_call,
                            inner=e))


callbacks = {
    dict: partial(_resolve_sequence,
                  item_coerce=lambda val: val.iteritems(),
                  container_coerce=dict),
    list: partial(_resolve_sequence, container_coerce=list),
    set: partial(_resolve_sequence, container_coerce=set),
    tuple: partial(_resolve_sequence, container_coerce=tuple),
}


def resolve_dict(dict_obj, extra_callbacks=None):
    """Resolves all keys in the given dictionary, following all Ref objects.

    Args:
      dict_obj - dict<string, object> A dictionary.
      extra_callbacks - dict<type, func> A dictionary mapping a specific type
        to a function used to resolve instances of that type. Mappings are
        provided for basic container objects.

    Returns:
      dict<string, object> - Resolved settings, following all Ref objects and
        any custom callbacks.
    """
    extra_callbacks = extra_callbacks or {}

    # Should be able to construct a DAG for dependencies.
    context = ResolutionContext(dict_obj, callbacks)
    for item_type, callback in extra_callbacks.iteritems():
        context.add_callback(item_type, callback)

    result = {}
    for key in dict_obj.iterkeys():
        result[key] = context.Get(key)

    return result


def resolve_from_object(obj, extra_callbacks=None):
    """Resolves the given instance, resolving all UPPER_CASED attribute names.

    This is similar to resolve_dict, but works on non-dictionary objects.
    Currently the convention is to examine only attributes which are
    UPPER_CASED. Private attributes (i.e. start with an '_' character) are not
    included.

    Args:
      obj - object Object whose attributes represent settings to be resolved.
      extra_callbacks - dict<type, func> A dictionary mapping a specific type
        to a function used to resolve instances of that type. Mappings are
        provided for basic container objects.

    Returns:
      dict<string, object> - Resolved settings, following all Ref object and
        any custom callbacks.
    """
    resolution_context = {name: getattr(obj, name)
                          for name in dir(obj)
                          if name == name.upper() and not name.startswith('_')}
    return resolve_dict(resolution_context, extra_callbacks=extra_callbacks)
