import os
import unittest

from collections import OrderedDict

from resolve import Ref, resolve_dict, resolve_from_object, Call


class ResolveDictCyclesTest(unittest.TestCase):

    def test_no_cycle(self):
        container = {
            'a': Ref('b'),
            'b': 1,
        }
        resolve_dict(container)

    def test_non_existent_reference(self):
        container = {'a': Ref('b')}
        with self.assertRaisesRegexp(ValueError,
                                     ("'b' could not be resolved in context "
                                      "\\['a'\\]")):
            resolve_dict(container)

    def test_mini_cycle(self):
        container = OrderedDict([
            ('a', Ref('b')),
            ('b', Ref('a')),
        ])

        with self.assertRaisesRegexp(ValueError,
                                     'Circular dependency found: a -> b -> a'):
            resolve_dict(container)

    def test_multiple_references_to_same_value(self):
        container = {
            'a': 1,
            'b': Ref('a') + 4 + Ref('a')
        }

        result = resolve_dict(container)

        self.assertEquals({
            'a': 1,
            'b': 6
        }, result)

    def test_large_cycle(self):
        container = OrderedDict([
            ('a', Ref('b')),
            ('b', Ref('c')),
            ('c', Ref('d')),
            ('d', Ref('a')),
        ])

        with self.assertRaisesRegexp(ValueError,
                                     ('Circular dependency found: '
                                      'a -> b -> c -> d -> a')):
            resolve_dict(container)

    def test_cycles_with_additional_nodes(self):
        container = OrderedDict([
            ('a', Ref('c')),
            ('b', 2),
            ('c', Ref('b') + Ref('b') + Ref('b') + Ref('a')),
        ])

        with self.assertRaisesRegexp(ValueError,
                                     ('Circular dependency found: '
                                      'a -> c -> a')):
            resolve_dict(container)

    def test_cycle_with_self(self):
        container = {'a': Ref('a')}

        with self.assertRaisesRegexp(ValueError,
                                     ('Circular dependency found: '
                                      'a -> a')):
            resolve_dict(container)

    def test_no_nodes(self):
        resolve_dict({})

    def test_cycle_in_list(self):
        container = {
            'a': Ref('b'),
            'b': [1, 2, 3, Ref('a')],
        }

        with self.assertRaisesRegexp(ValueError,
                                     ('Circular dependency found: '
                                      'a -> b -> a')):
            resolve_dict(container)

    def test_cycle_in_deeply_nested_dictionary(self):
        container = {
            'a': Ref('b'),
            'b': {'foo': {'bar': Ref('a')}},
        }

        with self.assertRaisesRegexp(ValueError,
                                     ('Circular dependency found: '
                                      'a -> b -> a')):
            resolve_dict(container)


class ResolveDictTest(unittest.TestCase):

    def test_primitive_value(self):
        container = {'foo': 1}
        result = resolve_dict(container)
        self.assertEquals({'foo': 1}, result)

    def test_ref_value(self):
        container = {
            'foo': 1,
            'bar': Ref('foo')
        }
        result = resolve_dict(container)

        self.assertEquals({
            'foo': 1,
            'bar': 1
        }, result)

    def test_chained_ref(self):
        container = {
            'a': 1,
            'b': Ref('a'),
            'c': Ref('b')
        }

        result = resolve_dict(container)

        self.assertEquals({
            'a': 1,
            'b': 1,
            'c': 1,
        }, result)

    def test_addition(self):
        container = {
            'a': 1,
            'b': Ref('a') + 100,
            'c': 200 + Ref('a'),
            'd': Ref('c') + Ref('b') + 300,
            'e': 400 + Ref('c') + Ref('b')
        }

        result = resolve_dict(container)

        self.assertEquals({
            'a': 1,
            'b': 101,
            'c': 201,
            'd': 602,
            'e': 702,
        }, result)

    def test_subtraction(self):
        container = {
            'a': 1,
            'b': Ref('a') - 100,
            'c': 200 - Ref('a'),
            'd': Ref('c') - Ref('b') - 300,
            'e': 400 - Ref('c') - Ref('b')
        }

        result = resolve_dict(container)

        self.assertEquals({
            'a': 1,
            'b': -99,
            'c': 199,
            'd': -2,
            'e': 300,
        }, result)

    def test_non_numeric_addition(self):
        container = {
            'a': 'this ',
            'b': 'is ',
            'c': Ref('a'),
            'd': Ref('b'),
            'e': ('I say ' + Ref('a') + Ref('b') + Ref('c')
                  + Ref('d') + 'fun!')
        }

        result = resolve_dict(container)

        self.assertEquals({
            'a': 'this ',
            'b': 'is ',
            'c': 'this ',
            'd': 'is ',
            'e': 'I say this is this is fun!'
        }, result)

    def test_unsupported_type_for_magic_method(self):
        container = {
            'a': 'this',
            'b': 2,
            'c': Ref('a') + Ref('b')
        }

        with self.assertRaisesRegexp(TypeError,
                                     ('Problem resolving call: '
                                      '\'this\' \\+ 2')):
            resolve_dict(container)

    def test_multiplication(self):
        container = {
            'a': '!',
            'b': 5,
            'c': Ref('a') * Ref('b')
        }

        result = resolve_dict(container)

        self.assertEquals({
            'a': '!',
            'b': 5,
            'c': '!!!!!',
        }, result)

    def test_nested_sequences_resolved(self):
        container = {
            'a': 1,
            'b': 2,
            'c': 3,
            'd': 4,
            'e': [5, 6, Ref('a'), Ref('b'), Ref('c') + Ref('d')],
            'f': {
                'g': Ref('a'),
                'h': Ref('e'),
            },
            'i': set([7, 8, Ref('a'), 9, Ref('b')])
        }

        result = resolve_dict(container)

        self.assertEquals({
            'a': 1,
            'b': 2,
            'c': 3,
            'd': 4,
            'e': [5, 6, 1, 2, 7],
            'f': {
                'g': 1,
                'h': [5, 6, 1, 2, 7],
            },
            'i': set([7, 8, 1, 9, 2])
        }, result)

    def test_call_ref(self):
        container = {
            'a': 'static',
            'b': 'filepath',
            'c': Call(os.path.join, '/tmp', Ref('a'), Ref('b')),
        }

        result = resolve_dict(container)

        self.assertEquals({
            'a': 'static',
            'b': 'filepath',
            'c': '/tmp/static/filepath',
        }, result)

    def test_call_with_kwargs(self):
        def some_func(foo, bar):
            return foo - bar

        container = {
            'a': 1,
            'b': 3,
            'c': Call(some_func, Ref('b'), bar=Ref('a'))
        }

        result = resolve_dict(container)

        self.assertEquals({
            'a': 1,
            'b': 3,
            'c': 2,
        }, result)

    def test_problem_calling_with_kwargs(self):
        def some_func(foo, bar):
            raise ValueError('Some error')

        container = {
            'a': 1,
            'b': 2,
            'c': Call(some_func, Ref('a'), bar=Ref('b'))
        }

        with self.assertRaisesRegexp(TypeError,
                                     ('Problem resolving call: '
                                      'some_func\\(1, bar=2\\).')):
            resolve_dict(container)

    def test_call_with_bad_references(self):
        container = {
            'a': Call(os.path.join, Ref('b')),
        }

        with self.assertRaisesRegexp(ValueError,
                                     ("'b' could not be resolved in context "
                                      "\\['a'\\]")):
            resolve_dict(container)

    def test_call_with_circular_reference(self):
        container = {
            'a': Ref('b'),
            'b': Call(os.path.join, Ref('a')),
        }

        with self.assertRaisesRegexp(ValueError,
                                     ('Circular dependency found: '
                                      'a -> b -> a')):
            resolve_dict(container)

    def test_call_with_nested_reference(self):
        container = {
            'a': 'static',
            'b': 'filepath',
            'c': 'someotherfilepath',
            'd': [
                Call(os.path.join, Ref('a'), Ref('b')),
                Call(os.path.join, Ref('a'), Ref('c')),
            ],
            'e': {
                'f': Call(map, lambda item: item + 'foo', [Ref('b'), Ref('c')])
            }
        }

        result = resolve_dict(container)

        self.assertEquals({
            'a': 'static',
            'b': 'filepath',
            'c': 'someotherfilepath',
            'd': ['static/filepath', 'static/someotherfilepath'],
            'e': {
                'f': ['filepathfoo', 'someotherfilepathfoo'],
            }
        }, result)

    def test_cannot_reference_non_top_level_item(self):
        container = {
            'a': {
                'b': 1,
            },
            'c': Ref('b'),
        }

        with self.assertRaisesRegexp(ValueError,
                                     ("'b' could not be resolved in context "
                                      "\\['a', 'c'\\]")):
            resolve_dict(container)

    def test_callback_override(self):
        container = {
            'a': [1, 2, 3],
        }

        extra_callbacks = {
            list: lambda r, context: 'foo'
        }

        result = resolve_dict(container, extra_callbacks=extra_callbacks)

        self.assertEquals({
            'a': 'foo'
        }, result)

        # Also make sure that the overrides are only valid for one resolution.
        result = resolve_dict(container)

        self.assertEquals({
            'a': [1, 2, 3],
        }, result)

    def test_callback_override_custom_type(self):
        class SomeSampleClass(object):

            def __init__(self, items):
                self.items = items

        container = {
            'a': SomeSampleClass([1, 2, 3])
        }
        extra_callbacks = {
            SomeSampleClass: lambda r, context: r.items
        }

        result = resolve_dict(container, extra_callbacks)

        self.assertEquals({
            'a': [1, 2, 3],
        }, result)


class ResolveFromObjectTest(unittest.TestCase):

    def test_inclusion(self):
        class Base(object):
            A = 1
            B = Ref('C')
            c = 'should_be_excluded'

            # Should be excluded.
            def some_method(self):
                pass

        class Prod(Base):
            C = 1
            D = Ref('A')
            E = 2
            _PRIVATE_ATTRIBUTE = 'should_be_excluded'

        result = resolve_from_object(Prod())

        self.assertEquals({
            'A': 1,
            'B': 1,
            'C': 1,
            'D': 1,
            'E': 2
        }, result)

if __name__ == '__main__':
    unittest.main()
